@@ -19,12 +19,14 @@
 """
 import logging 
 import base64
+from typing import List
 from hashlib import sha256
 from struct import pack, unpack
 from ecdsa.curves import SECP256k1
 from ecdsa.util import sigdecode_der
 
 from .ecc import ECPubkey, InvalidECPointException, sig_string_from_der_sig, sig_string_from_r_and_s, get_r_and_s_from_sig_string, CURVE_ORDER
+from .JCconstants import *
 
 logger = logging.getLogger(__name__)
 logger.setLevel(logging.DEBUG)
@@ -543,5 +545,185 @@ class CardDataParser:
             return False, "Invalid X9.62 encoding of the public key: " + bytes(pubkey).hex()
         
         
+    #################################
+    #                    SATODIME                 #        
+    #################################    
+    
+    def parse_satodime_get_keyslot_status(self, response: List[int]) -> dict:
+        
+        #response: [ key_status(1b) | key_type(1b) | key_asset(1b) | key_slip44(4b) | key_contract(34b) | key_tokenid(34b) | key_data(66b) ]
+        status_size= len(response)
+        if (status_size<SIZE_KEY_METADATA):
+            raise Exception(f"Satodime error: wrong keyslot status response length {status_size}, expected {SIZE_KEY_METADATA}")
+        offset=0
+        # unlock_code= response[offset:SIZE_UNLOCK_CODE] # deprecated
+        # offset+=SIZE_UNLOCK_CODE
+        key_status= response[offset]
+        offset+=1
+        key_type= response[offset]
+        offset+=1
+        key_asset= response[offset]
+        offset+=1
+        key_slip44=  response[offset:(offset+SIZE_SLIP44)]
+        offset+=SIZE_SLIP44
+        key_contract=  response[offset:(offset+SIZE_CONTRACT)]
+        offset+=SIZE_CONTRACT
+        key_tokenid=  response[offset:(offset+SIZE_TOKENID)]
+        offset+=SIZE_TOKENID
+        key_data=  response[offset:(offset+SIZE_DATA)]
+        offset+=SIZE_DATA
+        
+        # also convert to txt/hex values
+        #unlock_code_hex= bytes(unlock_code).hex()
+        key_status_txt=  DIC_STATE.get(key_status, f"Unknown code {key_status}")
+        key_asset_txt= DIC_ASSET_BY_CODE.get(key_asset, f"Unknown code {key_asset}")
+        key_slip44_hex= bytes(key_slip44).hex() # todo!
+        
+        size_contract= key_contract[1]
+        key_contract_hex= "0x"+ bytes(key_contract[2:(2+size_contract)]).hex() # parse as tlv
+        #key_contract_hex= bytes(key_contract).hex() # todo: parse as tlv
+        size_tokenid= key_tokenid[1]
+        key_tokenid_hex= "0x"+ bytes(key_tokenid[2:(2+size_tokenid)]).hex() # parse as tlv
+        #key_tokenid_hex= bytes(key_tokenid).hex() #todo
+        size_data= key_data[1] # key_data[0] is RFU (could be used as a type tag)
+        key_data_txt= bytes(key_data[2:(2+size_data)]).decode("utf-8")  #parse as tlv-encoded utf8 string
+        #key_data_txt= bytes(key_data).hex() #todo
+        
+        # token/nft
+        is_token= False
+        is_nft= False
+        if key_asset >=TOKEN_RANGE[0] and key_asset <TOKEN_RANGE[1]:
+            is_token= True
+        if key_asset >=NFT_RANGE[0] and key_asset <NFT_RANGE[1]:
+            is_nft= True
+        
+        keyslot_status={'key_status':key_status, 'key_type':key_type, 
+                                'key_asset':key_asset, 'key_slip44':key_slip44, 'key_contract':key_contract, 
+                                'key_tokenid':key_tokenid, 'key_data':key_data,
+                                'key_status_txt':key_status_txt, 
+                                'key_asset_txt':key_asset_txt, 'key_slip44_hex':key_slip44_hex,
+                                'key_contract_hex':key_contract_hex, 'key_tokenid_hex':key_tokenid_hex,
+                                'key_data_txt':key_data_txt, 'is_token':is_token, 'is_nft':is_nft}
+        
+        return keyslot_status
+        
+    def parse_satodime_get_pubkey(self, response):
+    
+        # response= [ pubkey_size(2b) | pubkey | sig_size(2b) | sig ]
+        response_size= len(response)
+        offset=0
+        remain= response_size
+
+        # pubkey
+        if (remain<2):
+            raise Exception(f"Satodime error: wrong get_pubkey response length {response_size}, expected at least {2}")
+        pubkey_size= response[0]*256 + response[1]
+        offset+=2
+        remain-=2
+        if (remain<pubkey_size):
+            raise Exception(f"Satodime error: wrong get_pubkey response length {response_size}, expected at least {2+pubkey_size}")
+        pubkey_list= response[offset:(offset+pubkey_size)]
+        offset+=pubkey_size
+        remain-=pubkey_size
+        # sig
+        if (remain<2):
+            raise Exception(f"Satodime error: wrong get_pubkey response length {response_size}, expected at least {2+pubkey_size+2}")
+        sig_size= response[offset]*256 + response[offset+1]
+        offset+=2
+        remain-=2
+        if (remain<sig_size):
+            raise Exception(f"Satodime error: wrong get_pubkey response length {response_size}, expected at least {2+pubkey_size+2+sig_size}")
+        sig_list= response[offset:(offset+sig_size)]
+        offset+=sig_size
+        remain-=sig_size
+        
+        # check signature
+        pubkey_hex= bytes(pubkey_list).hex()
+        logger.debug(f"Verifying sig for pubkey {pubkey_hex} using authentikey {self.authentikey.get_public_key_bytes(compressed=False).hex()}")
+        self.verify_signature(response[0:(2+pubkey_size)], sig_list, self.authentikey)
+        
+        # compressed key
+        parity= pubkey_list[64]%2
+        pubkey_comp_list= pubkey_list[0:33]
+        pubkey_comp_list[0]= 0x02 if (parity==0) else 0x03
+        return (pubkey_list, pubkey_comp_list, sig_list)
+        
+    def parse_satodime_get_privkey(self, response, key_nbr):
+    
+        #return: [ entropy_size(2b) | user_entropy + authentikey_coordx + card_entropy | privkey_size(2b) | privkey | sig_size(2b) | sig ]
+        response_size= len(response)
+        offset=0
+        remain= response_size
+        
+        # entropy_size
+        if (remain<2):
+            raise Exception(f"Satodime error: wrong get_privkey response length {response_size}, expected at least {2}")
+        entropy_size= response[0]*256 + response[1]
+        offset+=2
+        remain-=2
+        if (remain<entropy_size):
+            raise Exception(f"Satodime error: wrong get_privkey response length {response_size}, expected at least {2+entropy_size}")
+        entropy_list= response[offset:(offset+SIZE_ENTROPY+SIZE_ECCOORDX+SIZE_ENTROPY)]
+        entropy_user= response[offset:(offset+SIZE_ENTROPY)]
+        offset+=SIZE_ENTROPY
+        remain-=SIZE_ENTROPY
+        entropy_coordx= response[offset:(offset+SIZE_ECCOORDX)]
+        offset+=SIZE_ECCOORDX
+        remain-=SIZE_ECCOORDX
+        entropy_card= response[offset:(offset+SIZE_ENTROPY)]
+        offset+=SIZE_ENTROPY
+        remain-=SIZE_ENTROPY
+        
+        #privkey
+        if (remain<2):
+            raise Exception(f"Satodime error: wrong get_privkey response length {response_size}, expected at least {2+entropy_size+2}")
+        privkey_size= response[offset]*256 + response[offset+1]
+        offset+=2
+        remain-=2
+        if (remain<privkey_size):
+            raise Exception(f"Satodime error: wrong get_privkey response length {response_size}, expected at least {2+entropy_size+2+privkey_size}")
+        privkey_list= response[offset:(offset+privkey_size)]
+        #privkey_hex= bytes(privkey_list).hex()
+        offset+=privkey_size
+        remain-=privkey_size
+        # sig
+        if (remain<2):
+            raise Exception(f"Satodime error: wrong get_privkey response length {response_size}, expected at least {2+entropy_size+2+privkey_size+2}")
+        sig_size= response[offset]*256 + response[offset+1]
+        offset+=2
+        remain-=2
+        if (remain<sig_size):
+            raise Exception(f"Satodime error: wrong get_privkey response length {response_size}, expected at least {2+entropy_size+2+privkey_size+2+sig_size}")
+        sig_list= response[offset:(offset+sig_size)]
+        offset+=sig_size
+        remain-=sig_size
+        
+        # check signature
+        self.verify_signature(response[0:(2+entropy_size+2+privkey_size)], sig_list, self.authentikey)
+        
+        # check that entropy_coordx correspond to authentikey_coordx
+        if (entropy_coordx != self.authentikey_coordx):
+            raise Exception(f"Satodime error: entropy_coordx {bytes(entropy_coordx).hex()} for keyslot {key_nbr} does not match authentikey_coordx {bytes(authentikey_coordx).hex()}")
+        
+        # check that privkey is correctly derived from entropy:
+        digest= sha256()
+        digest.update(bytes(entropy_list))
+        entropy_hash_bytes=digest.digest()
+        if (list(entropy_hash_bytes) != privkey_list):
+            logger.debug(f"####### DEBUG ENTROPY => PRIVKEY #######")
+            logger.debug(f"entropy_list: {bytes(entropy_list).hex()}")
+            logger.debug(f"entropy_hash_bytes: {bytes(entropy_hash_bytes).hex()}")
+            logger.debug(f"privkey_list: {bytes(privkey_list).hex()}")
+            logger.debug(f"####### ######################### #######")
+            raise Exception(f"Satodime error: recovered private key for keyslot {key_nbr} does not match entropy!")
+        
+        return (entropy_list, privkey_list, sig_list)
+        
+        
+        
+        
+        
+        
+        
         
     
\ No newline at end of file
